/*
 ============================================================================
 Name        : lex_make.c
 Author      : H. Carl Ott
 Version     : see below
 Copyright   : JAN-2018
 Description : Read config text file, process text files and generate zipped .app, .dat, and .json files for RPTR-OLED OTA

 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <unistd.h>
#include "miniz_zip.h"

#define	MY_VERSION			"LEXICON MAKE UTIL VER NRF1.2e,  Jan 2018 HCO\n"

#define MANIFEST_FILENAME	"manifest.json"
#define DAT_FILE_SIZE		14			// bytes
#define MANIFEST_MAX_SIZE	1000

#define MAX_APP_FILE_SIZE	 0x23c00	// 146,432

// constants for .dat and manifest file, not enforced by bootloader
#define	APPLICATION_VERSION		0xffffffffUL
#define DEVICE_REVISION			0xffffUL
#define	DEVICE_TYPE				0xffffUL
#define	SOFTDEVICE_REQ			0xfffeUL
#define NUM_SOFTDEVICES			01UL

// data locations in output app file
#define	FLASH_OFFSET		0x18000
#define	LEX_CFG_OFFSET		0xA800			// fixed offset to cfg struct in the output file
#define	LEX_WORDS_OFFSET	0xA840			// fixed offset to word storage in the output file

#define LEX_CFG_SIG			0xAA55			// marker

#define MAX_WORD_FILES 8
#define CONFIG_LINES (3 + MAX_WORD_FILES)
#define	WORD_BUFFER_SIZE	256000			// ram allocated for processed slw and flw buffers
#define WORD_DELIMS			" ,."			// valid word delims

#define byte_0(value) 	(uint8_t) ((value & 0x000000FF) >> 0)
#define byte_1(value) 	(uint8_t) ((value & 0x0000FF00) >> 8)
#define byte_2(value)	(uint8_t) ((value & 0x00FF0000) >> 16)
#define byte_3(value)   (uint8_t) ((value & 0xFF000000) >> 24)


/**
 * lexicon config struct,
 * shared with rptr-oled app
 * revised cnts to 32bits to align properly
 */
typedef struct{
	uint16_t	lex_sig;		// valid sig (not used)
	uint8_t  	name[6];		// lexicon name
	uint32_t	flw_offset;		// FLW data offset
	uint32_t	flw_cnt;		// number of FLW
	uint32_t	slw_offset;   	// SLW data offset
	uint32_t	slw_cnt;		// number of SLW
}lex_cfg_t;


/**
 * Prototypes
 */

int	process_file(FILE * file);
int	fill_dat_file_buff(char * dat_buf, unsigned int crc);
int	fill_manifest_buff(char * buff, char * bin_file,  char * dat_file, unsigned int crc);
int	fill_lex_cfg(char *buf, lex_cfg_t * cfg);


/**
 * globals
 */
int 	line_cnt = 0;								// stats
int		slw_cnt	 = 0;								//
int		flw_cnt	= 0;								//
int		slw_dirty_cnt = 0;						    //
int		flw_dirty_cnt = 0;							//

char	*slw_idx;									// index for SLWs
char	*flw_idx;									// index for FLWs

char	*pslw_buff;									// pointers to word buffers
char	*pflw_buff;

char	*pman_buff;									// pointer to manifest buff

char	*papp_buff;									// pointer to app buff
char	*app_idx;									// app write idx

char	dat_file_buff[DAT_FILE_SIZE];

/**
 * main
 */

int main(int argc, char *argv[]) {
	char  	line_buff[255];								// gp line buffer

	int		i = 0;										// gp counter
	int		warnings = 0;
	long	sz;
	char	cfg_line[CONFIG_LINES][255];			    // config file text file copied to array
	char	dat_filename[255];							// .dat file name
	char	app_filename[255];							// filename of app bin file
	char	pkg_filename[255];							// base filename of pkg
	char	zip_filename[255];							// zip filename

	char	lex_name[7];								// size of lex + trailing NULL

	unsigned int	file_crc;							// crc - 16 = CCITT of new bin file
	lex_cfg_t		lex_cfg;							// lexicon config struct,  let RPTR-OLED app know where data is located

	int		ret_val;									// gp

	mz_bool status;										// miniz status
	mz_zip_archive zip_archive;

	memset(line_buff,0, sizeof(line_buff));				// init line buffer
	memset(cfg_line, 0, sizeof(cfg_line));				// init config array

	FILE	*cfg_file;									// file handles
	FILE	*inp_bin_file;
	FILE	*wrd_file;

	int	out_file_size;

	printf(MY_VERSION);

	/* check for proper number of cmd line args */
	if (argc != 2) {
		printf("ERROR: Wrong number of arguments. Usage \"%s config_file\"\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	/* try to open config file */
	cfg_file = fopen(argv[1], "r");

	if (cfg_file == NULL) {
		printf("ERROR: Can't open config file: \"%s\".\n", argv[1]);
		exit(EXIT_FAILURE);
	}

	printf("Processing config file:\"%s\"\n", argv[1]);

	i = 0;

	while ( fgets(line_buff, sizeof(line_buff), cfg_file) != NULL)	{		// read line from file
		line_buff[strcspn(line_buff, "\n" )] = '\0';						// convert the newline to null
		line_buff[strcspn(line_buff, "\r" )] = '\0';						// convert the cr to null

		if (i < CONFIG_LINES && strlen(line_buff) > 0 && strlen(line_buff) < sizeof(line_buff) - 1 ) {
			strcpy(cfg_line[i++], line_buff);							// copy to config array
		}

		else{
			printf("WARNING: Config line %d discarded.\n", i+1);
			i++;
			warnings++;
			if (warnings > 15){
				printf("ERROR: 16 or more warnings. Check your config file.\n");
				exit(EXIT_FAILURE);
			}

		}

	}

	if (cfg_file){
		fclose(cfg_file);	// all done with cfg file
	}

	// get output package, remove any extension

	strcpy(pkg_filename, cfg_line[1]);
	pkg_filename[strcspn(pkg_filename, ".")]  = '\0';

	/* make output file names */
	strcpy(dat_filename, pkg_filename);
	strcpy(zip_filename, pkg_filename);
	strcpy(app_filename, pkg_filename);

	strcat(dat_filename, ".dat");
	strcat(zip_filename, ".zip");
	strcat(app_filename, ".app");

	/* malloc storage for words and manifest file */
	pslw_buff = malloc(WORD_BUFFER_SIZE);

	if (pslw_buff) {
		slw_idx = pslw_buff;
	}

	else {
		printf("ERROR: slw buff malloc failed.\n");		// error on malloc
		exit(EXIT_FAILURE);
	}

	pflw_buff = malloc(WORD_BUFFER_SIZE);

	if (pflw_buff) {
		flw_idx = pflw_buff;
	}

	else {
		printf("ERROR: flw buff malloc failed.\n");		// error on malloc
		exit(EXIT_FAILURE);
	}

	pman_buff = malloc(MANIFEST_MAX_SIZE);

	if (!pman_buff) {
		printf("ERROR: manifest file buffer malloc failed.\n");		// error on malloc
		exit(EXIT_FAILURE);
	}

	papp_buff = malloc(MAX_APP_FILE_SIZE);

	if (!papp_buff) {
		printf("ERROR: app file buffer malloc failed.\n");		// error on malloc
		exit(EXIT_FAILURE);
	}


	for (i = 3; i < CONFIG_LINES; ++i) {			// process word source files

		if( *cfg_line[i] != 0x0){					// make sure line is not empty

			wrd_file = 	fopen(cfg_line[i], "r");	// open file

			if (wrd_file){						 	// see if opened ok
				fseek(wrd_file, 0L, SEEK_END);
				sz = ftell(wrd_file);
				if (sz > WORD_BUFFER_SIZE){
					printf("ERROR: Text file \"%s\" too large (%ld bytes, limit %d bytes).\n", cfg_line[i], sz, WORD_BUFFER_SIZE);
					exit(EXIT_FAILURE);
				}

				fseek(wrd_file, 0L, SEEK_SET);
				ret_val = process_file(wrd_file);	// process the file
				printf("Processed \"%s\", %d total words were processed from text file.\n", cfg_line[i], ret_val);
				fclose(wrd_file);
			}

			else{
				printf("ERROR: Can't open \"%s\" for processing.\n", cfg_line[i]);		// error on file open
				exit(EXIT_FAILURE);
			}

		}

	}


	/* print some stats to console */
	printf("Total lines processed = %d\n",line_cnt);
	printf("Total regular four letter words = %d\n",flw_cnt);
	printf("Total regular six letter words = %d\n",slw_cnt);
	printf("Total dirty four letter words = %d\n",flw_dirty_cnt);
	printf("Total dirty six letter words = %d\n",slw_dirty_cnt);
	printf("Total words = %d\n",slw_dirty_cnt + flw_dirty_cnt + flw_cnt + slw_cnt);

	printf("Now generating output app data.\n");

	/* get source .bin code */
	inp_bin_file = fopen(cfg_line[0], "rb");

	if (inp_bin_file == NULL) {
		fprintf(stderr, "ERROR, Can't open RPTR-OLED input bin file: \"%s\".\n", cfg_line[0]);
		exit(EXIT_FAILURE);
	}

	if ( fread(papp_buff, 1, LEX_CFG_OFFSET, inp_bin_file ) < LEX_CFG_OFFSET ){
		fprintf(stderr, "ERROR: Input bin file was too short.\n");
		exit(EXIT_FAILURE);
	}

	fclose(inp_bin_file);									// done with input .bin file
	printf("%d bytes were read from input bin file into app buffer.\n", LEX_CFG_OFFSET);

	memcpy(lex_name, cfg_line[2], 6);				// copyuser-provided lexicon name
	lex_name[6] = '\0';								// enforce trailing null
	for (i = 0; i < 6; i++) {
		lex_name[i] = toupper(lex_name[i]);			// upper case letters
		if (!isalnum(lex_name[i])){
			lex_name[i] = ' ';						// allow only alphanumerics
		}
	}
	for (i = 0; i < 6; i++) {						// change NULL to space so we don't compain about short names
		if (cfg_line[2][i] == '\0'){
			cfg_line[2][i] = ' ';
		}
	}

	if (memcmp(lex_name, cfg_line[2], 6) != 0){
		printf("WARNING: Lexicon name was changed from \"%s\" to \"%s\".\n", cfg_line[2], lex_name);
	}

	memcpy(lex_cfg.name, lex_name, sizeof(lex_cfg.name));		// copy the lexicon name
	lex_cfg.slw_cnt = slw_cnt + slw_dirty_cnt;					// get the counts and offsets
	lex_cfg.flw_cnt = flw_cnt + flw_dirty_cnt;
	lex_cfg.slw_offset = FLASH_OFFSET + LEX_WORDS_OFFSET + 0;	 // slw first to get a 4 byte alignment
	lex_cfg.flw_offset = FLASH_OFFSET + LEX_WORDS_OFFSET + ((slw_cnt + slw_dirty_cnt) * 4);		// flw are right after the slw

	lex_cfg.lex_sig = LEX_CFG_SIG;					// set valid cfg signature

	out_file_size = LEX_WORDS_OFFSET + ( (slw_cnt + slw_dirty_cnt) * 4) + ( (flw_cnt + flw_dirty_cnt) * 3);

	// check for 4 byte alignment, otherwise dfu gives a very unhelpful error msg and just fails
	if ( out_file_size & 0x03 ){
		i = 4 - (out_file_size % 4);							// calc num bytes to pad
		while (i--){
			*(papp_buff + out_file_size++) = 0xff;				// pad with ff so we can see it in output, and adjist to new file size
		}
	}

	if (out_file_size > MAX_APP_FILE_SIZE ){
		printf("ERROR: The generated file is too large to fit in RPTR.\n");
		printf("Max file size is %d bytes.\n", MAX_APP_FILE_SIZE);
		exit(EXIT_FAILURE);
	}


	fill_lex_cfg(papp_buff + LEX_CFG_OFFSET, &lex_cfg );
	memcpy(papp_buff + LEX_WORDS_OFFSET, pslw_buff, (slw_cnt + slw_dirty_cnt) * 4 );
	memcpy(papp_buff + LEX_WORDS_OFFSET + ( (slw_cnt + slw_dirty_cnt) * 4), pflw_buff, (flw_cnt + flw_dirty_cnt) * 3 );

	/* init crc */

	app_idx = papp_buff;		// set index to buff

	file_crc = 0xffff;

	for (i = 0; i < out_file_size; i++) {					// generate crc for data

		file_crc = (file_crc >> 8 & 0x00FF) | (file_crc << 8 & 0xFF00);
		file_crc ^= (uint8_t) *app_idx++;

		file_crc ^= (file_crc & 0x00FF) >> 4;
		file_crc ^= (file_crc << 8) << 4;
		file_crc ^= ((file_crc & 0x00FF) << 4) << 1;
		file_crc &= 0xffff;
	}

	printf("Lexicon \"%s\" output to app buffer \"%s\".\n",lex_name, app_filename );
	printf("%d (0x%X) bytes were written to output app buffer with a generated crc16 of %d (0x%04X).\n", (int) out_file_size, (int) out_file_size, file_crc, file_crc);

	printf("Creating final zip output from temp data.\n");

	/* remove any existing old file */

	if( access( zip_filename, F_OK ) != -1 ) {
		printf("Deleting existing \"%s\" file.\n", zip_filename);
		remove(zip_filename);
	}

	memset(&zip_archive, 0, sizeof( zip_archive ) ); // init miniz to write files
	status = mz_zip_writer_init_file(&zip_archive, zip_filename, 0 );

	if (!status)  {
		printf("mz_zip_writer_init_file failed!\n");
		return EXIT_FAILURE;
	}

	// populate and write .dat data
	fill_dat_file_buff(dat_file_buff, file_crc);
	status = mz_zip_writer_add_mem(&zip_archive, dat_filename, &dat_file_buff, sizeof(dat_file_buff),MZ_BEST_COMPRESSION );

	if (!status)  {
		printf("mz_zip add .dat failed failed!\n");
		return EXIT_FAILURE;
	}


	// create manifest
	status = mz_zip_writer_add_mem(&zip_archive, MANIFEST_FILENAME, pman_buff, fill_manifest_buff(pman_buff, app_filename, dat_filename, file_crc), MZ_BEST_COMPRESSION);

	if (!status)  {
		printf("mz_zip add manifest failed!\n");
		return EXIT_FAILURE;
	}

	// write app file
	status = mz_zip_writer_add_mem(&zip_archive, app_filename, papp_buff, out_file_size, MZ_BEST_COMPRESSION);

	if (!status)  {
		printf("mz_zip add rptr app file failed!\n");
		return EXIT_FAILURE;
	}

	/* finalize the archive */
	mz_zip_writer_finalize_archive(&zip_archive);
	mz_zip_writer_end(&zip_archive);

	printf("Created \"%s\" OTA DFU package.\n", zip_filename);

	return EXIT_SUCCESS;
}


/**
 * write cfg data to buffer memory
 */

int	fill_lex_cfg( char *buf, lex_cfg_t * cfg){

	buf[0] = byte_0(cfg->lex_sig);
	buf[1] = byte_1(cfg->lex_sig);

    /* 2 - 7 */
	memcpy(buf + 2, cfg->name, sizeof(cfg->name) );

	buf[8] = byte_0(cfg->flw_offset);
	buf[9] = byte_1(cfg->flw_offset);
	buf[10] = byte_2(cfg->flw_offset);
	buf[11] = byte_3(cfg->flw_offset);

	buf[12] = byte_0(cfg->flw_cnt);
	buf[13] = byte_1(cfg->flw_cnt);
	buf[14] = byte_2(cfg->flw_cnt);
	buf[15] = byte_3(cfg->flw_cnt);

	buf[16] = byte_0(cfg->slw_offset);
	buf[17] = byte_1(cfg->slw_offset);
	buf[18] = byte_2(cfg->slw_offset);
	buf[19] = byte_3(cfg->slw_offset);

	buf[20] = byte_0(cfg->slw_cnt);
	buf[21] = byte_1(cfg->slw_cnt);
	buf[22] = byte_2(cfg->slw_cnt);
	buf[23] = byte_3(cfg->slw_cnt);

	return 0;
}



/**
 * populate the json manifest data to keep the DFU util happy
 */

int	fill_manifest_buff(char * buff, char * bin_file,  char * dat_file, unsigned int crc){

	int cnt;

	cnt =  sprintf(buff, "{\r\n");
	cnt += sprintf(buff + cnt, "    \"manifest\": {\r\n");
	cnt += sprintf(buff + cnt, "        \"application\": {\r\n");
	cnt += sprintf(buff + cnt, "            \"bin_file\": \"%s\",\r\n", bin_file);
	cnt += sprintf(buff + cnt, "            \"dat_file\": \"%s\",\r\n", dat_file);
	cnt += sprintf(buff + cnt, "            \"init_packet_data\": {\r\n");
	cnt += sprintf(buff + cnt, "                \"application_version\": %lu,\r\n", APPLICATION_VERSION);
	cnt += sprintf(buff + cnt, "                \"device_revision\": %lu,\r\n", DEVICE_REVISION);
	cnt += sprintf(buff + cnt, "                \"device_type\": %lu,\r\n", DEVICE_TYPE);
	cnt += sprintf(buff + cnt, "                \"firmware_crc16\": %d,\r\n", crc);
	cnt += sprintf(buff + cnt, "                \"softdevice_req\": [\r\n");
	cnt += sprintf(buff + cnt, "                    %lu\r\n", SOFTDEVICE_REQ );
	cnt += sprintf(buff + cnt, "                ]\r\n");
	cnt += sprintf(buff + cnt, "            }\r\n");
	cnt += sprintf(buff + cnt, "        },\r\n");
	cnt += sprintf(buff + cnt, "        \"dfu_version\": 0.5\r\n");
	cnt += sprintf(buff + cnt, "    }\r\n");
	cnt += sprintf(buff + cnt, "}");


	//printf("data :\r\n %s", buff);

	return cnt;
}


/**
 * fill dat file buffer with app data
 * most values not used by nootloader
 */

int	fill_dat_file_buff(char * dat_buf, unsigned int crc){

	// device type
	dat_buf[0] = byte_0(DEVICE_TYPE);
	dat_buf[1] = byte_1(DEVICE_TYPE);

	// device version
	dat_buf[2] = byte_0(DEVICE_REVISION);
	dat_buf[3] = byte_1(DEVICE_REVISION);

	// application version
	dat_buf[4] = byte_0(APPLICATION_VERSION);
	dat_buf[5] = byte_1(APPLICATION_VERSION);
	dat_buf[6] = byte_2(APPLICATION_VERSION);
	dat_buf[7] = byte_3(APPLICATION_VERSION);

	// num of softdevices in table
	dat_buf[8] = byte_0(NUM_SOFTDEVICES);
	dat_buf[9] = byte_1(NUM_SOFTDEVICES);

	// softdevices required
	dat_buf[10] = byte_0(SOFTDEVICE_REQ);
	dat_buf[11] = byte_1(SOFTDEVICE_REQ);

	dat_buf[12] = byte_0(crc);
	dat_buf[13]	= byte_1(crc);

	return 0;
}



/**
 * read word source text file,
 * parse and pack SLW and FLWs into ram buffers, update counts
 * any word that has at least one capitalized letter is flagged as a dirty word.
 */

int	process_file(FILE * file){

	char  	line_buff[255];								// gp line buffer
	char  	word_buff[16];								// word buffer
	char 	*pword;										// word pointer
	char	*pchar;										// char pointer
	int	 	dirty_flag = 0;								// was this a naughty word?
	unsigned int word_packed;							// place to build packed word  (minimum 32 bits)
	int		words = 0;

	if (file == NULL){									// check for valid fp
		return 0;
	}

	memset(line_buff,0,sizeof(line_buff));				// init line buffer

	while ( fgets(line_buff, sizeof(line_buff), file) != NULL){				// grab a full line


		line_buff[strcspn(line_buff, "\n" )] = '\0';						// convert the newline to zero terminated
		line_buff[strcspn(line_buff, "\r" )] = '\0';						// convert the cr to zero terminated
		line_cnt++;															// update line count stats

		if (strncmp ("//", line_buff, 2) == 0){								// if line starts with "//" treat as comment and discard
			continue;														// loop for next line
		}

		pword = strtok(line_buff,WORD_DELIMS);								// tokenize line buff

		while (pword != NULL){												// loop through each word on line

			strcpy(word_buff, pword);										// get the word into work buff
			dirty_flag = 0;													// clear flag
			word_packed = 0;										   		// init packed workspace

			/* process SLW */
			if  (strlen(word_buff) == 6){

				for (pchar = word_buff; *pchar; ++pchar){

					if(isupper(*pchar) ) {									// if any letter is caps, dirty
						dirty_flag = 1;
					}

					word_packed <<= 5;										// make room
					word_packed |= ((*pchar - 0x61) & 0x1f);				// or in new letter
				}

				word_packed <<= 2;											// now shift to make room for the two flag bits

				if( (slw_idx + 4) < (pslw_buff + WORD_BUFFER_SIZE)  ){				// check buffer for room

					/* see if dirty */
					if	(dirty_flag){										// copy dirty flag to lsb
						word_packed |= 0x01;
						slw_dirty_cnt++;									// bump stats
					}

					else{
						slw_cnt++;
					}

					*slw_idx++ = (word_packed >> 24) & 0xff;				// write to buffer
					*slw_idx++ = (word_packed >> 16) & 0xff;
					*slw_idx++ = (word_packed >> 8) & 0xff;
					*slw_idx++ = word_packed & 0xff;
					words++;
				}

				/* packed word won't fit in buffer, discard */
				else{
				}

			}

			/* process FLW */
			else if  (strlen(word_buff) == 4){



				for (pchar = word_buff; *pchar;  ++pchar){

					if(isupper(*pchar) ) {							// if any letter is caps, dirty
						dirty_flag = 1;
					}

					word_packed <<= 5;								// make room for new letter
					word_packed |= ((*pchar - 0x61) & 0x1f);		// or in new letter
				}

				word_packed <<= 4;									//  shift for flag bits

				if( (flw_idx + 3) < (pflw_buff + WORD_BUFFER_SIZE) ){	// check buffer for rimm


					if	(dirty_flag){								// copy dirty flag to lsb
						word_packed |= 0x01;
						flw_dirty_cnt++;							// bump stats
					}

					else{
						flw_cnt++;
					}
					*flw_idx++ = (word_packed >> 16) & 0xff;
					*flw_idx++ = (word_packed >> 8) & 0xff;
					*flw_idx++ = word_packed & 0xff;
					words++;
					// bump stats
				}

				/* word won't fit in buffer */
				else{
				}


			}


			/* process other length words */
			else{			// wrong word length, ignore it

			}

			pword = strtok(NULL, WORD_DELIMS);			// advance to next word in line buff

		}


	}

	return words;

}

